// Q's C++ Library
// Copyright (c) 1998-1999 Hisashi HORIKAWA. All rights reserved.

#ifdef DEBUG

#include <cstdlib>
#include <cstdio>
#include <list>
#include <string>
#include <set>
#include <algorithm>
using namespace std;

//////////////////////////////////////////////////////////////////////
// TRACE

static bool trace_use_file = false;
static string trace_filename;

void q_trace(const char* format, ...)
{
    va_list args;
    va_start(args, format);
    // g_logv(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, format, args);
            // g_log()˲Ԥ
    if (!trace_use_file)
        vprintf(format, args);
    else {
        FILE* fp = fopen(trace_filename.c_str(), "a");
        if (fp) {
            vfprintf(fp, format, args);
            fclose(fp);
        }
    }
    va_end(args);
}

void q_trace_set_file(const char* fname)
{
    assert(fname && *fname);
    trace_filename = fname;
    trace_use_file = true;
}

//////////////////////////////////////////////////////////////////////
// LeakChecker

struct MallocData
{
    void* ptr;
    size_t size;
    string filename;
    int line;
    int type;

    bool operator == (const MallocData& a) const {
        return ptr == a.ptr;
    }
};

typedef list<MallocData> MallocList;
MallocList* pointers = 0;

class LeakChecker
{
public:
    LeakChecker() {
        atexit(check);
        pointers = new MallocList();
    }
private:
    static void check();
    static void str_out(unsigned char* ptr, size_t size);
} leakChecker;

void LeakChecker::str_out(unsigned char* ptr, size_t size)
{
    assert(ptr);

    char buf[1000];
    char* p = buf;
    for (unsigned i = 0; i < size && i < 16; i++)
        p += sprintf(p, "%02x ", ptr[i]);
    q_trace("%s'", buf);

    p = buf;
    for (unsigned i = 0; i < size && i < 16; i++) {
        if (*ptr >= 0x20 && *ptr <= 0x7e)
            *p++ = *ptr++;
        else {
            ptr++;
            *p++ = '.';
        }
        if (p - buf >= sizeof(buf) - 1) {
            *p++ = 0;
            q_trace("%s", buf);
            p = buf;
        }
    }
    if (p - buf > 0) {
        *p++ = 0;
        q_trace("%s'\n", buf);
    }
    else
        q_trace("'\n");
}

struct MallocDataComp {
    bool operator () (const MallocData* x, const MallocData* y) const {
        if (x->filename != y->filename)
            return x->filename < y->filename;
        return x->line < y->line;
}};

typedef set<MallocData*, MallocDataComp> MallocSet;
MallocSet* mset = 0;
    // ƱԤΥ꡼̤ɽȡäݵƫ
    // setꤲơ1ԤФ

void LeakChecker::check()
{
    if (pointers->size()) {
        q_trace("warning: memory leak = %d\n", pointers->size());
        mset = new MallocSet();

        MallocList::iterator i;
        for (i = pointers->begin(); i != pointers->end(); i++)
            mset->insert(&*i);

        MallocSet::iterator it;
        for (it = mset->begin(); it != mset->end(); it++) {
            q_trace("%s(%d): ", (*it)->filename.c_str(), (*it)->line);
            str_out((unsigned char*) (*it)->ptr, (*it)->size);
        }
    }
    // delete pointers;
        // TODO: ꡼å꡼򵯤Τ򤹤
}

//////////////////////////////////////////////////////////////////////
// for plain C

static void q_malloc_sub(void* ptr, size_t size, const char* filename, int line, int type)
{
    MallocData data;
    data.ptr = ptr;
    data.size = size;
    data.filename = filename;
    data.line = line;
    data.type = type;
    pointers->push_back(data);
}

void* q_malloc(size_t size, const char* filename, int line)
{
    void* r = malloc(size);
    if (r)
        q_malloc_sub(r, size, filename, line, 1);
    else
        perror("malloc failed");
    return r;
}

void* q_calloc(size_t nelem, size_t elsize, const char* filename, int line)
{
    void* r = calloc(nelem, elsize);
    if (r)
        q_malloc_sub(r, nelem * elsize, filename, line, 1);
    else
        perror("calloc");
    return r;
}

void* q_realloc(void* ptr, size_t size, const char* filename, int line)
{
    MallocData data;
    data.ptr = ptr;
    MallocList::iterator i = find(pointers->begin(), pointers->end(), data);
    if (i != pointers->end()) {
        if (i->type != 1) {
            q_trace("malloc type is not match\n");
            assert(0);
        }
        pointers->erase(i);
    }
    else {
        q_trace("warning: realloc(): invalid ptr\n");
    }
    void* r = realloc(ptr, size);
    if (r)
        q_malloc_sub(r, size, filename, line, 1);
    else
        perror("realloc");
    return r;
}

void* q_valloc(size_t size, const char* filename, int line)
{
    void* r = valloc(size);
    if (r)
        q_malloc_sub(r, size, filename, line, 1);
    else
        perror("valloc");
    return r;
}

static void q_free_sub(void* ptr, const char* filename, int line, int type)
{
    if (ptr) {
        MallocData data;
        data.ptr = ptr;
        MallocList::iterator i = find(pointers->begin(), pointers->end(), data);
        if (i != pointers->end()) {
            if (i->type != type) {
                q_trace("type mismatch\n");
                assert(0);
            }
            pointers->erase(i);
        }
        else {
            if (filename && *filename)
                q_trace("%s(%d): free(): invalid ptr\n", filename, line);
        }
        free(ptr);
    }
}

void q_free(void* ptr, const char* filename, int line)
{
    q_free_sub(ptr, filename, line, 1);
}

char* q_strdup(const char* s1, const char* filename, int line)
{
    assert(s1);
    char* ret = (char*) q_malloc(strlen(s1) + 1, filename, line);
    strcpy(ret, s1);
    return ret;
}

//////////////////////////////////////////////////////////////////////
// for C++

void* operator new(size_t size, const char* filename, int line) throw(bad_alloc)
{
    void* ptr = malloc(size);
    if (ptr)
        q_malloc_sub(ptr, size, filename, line, 2);
    else
        perror("new");
    return ptr;
}

void* operator new [] (size_t size, const char* filename, int line) throw(bad_alloc)
{
    void* ptr = malloc(size);
    if (ptr)
        q_malloc_sub(ptr, size, filename, line, 3);
    else
        perror("new []");
    return ptr;
}

void operator delete(void* p) throw()
{
    q_free_sub(p, 0, 0, 2);
}

void operator delete [] (void* p) throw()
{
    q_free_sub(p, 0, 0, 3);
}

#endif DEBUG
